"""Mu serialization and formatting helpers.

This module provides `json.dumps`-style APIs for rendering Python values and
Mu AST nodes to Mu source text in either concise or pretty form.
"""

from __future__ import annotations

import dataclasses
import re
import typing
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Annotated, Any, get_args, get_origin

from mu.quoted import Quoted
from mu.typed import FieldName, OneOrMore, OptionalArg, ZeroOrMore
from mu.types import (
    AtomExpr,
    Document,
    Expr,
    GroupExpr,
    MappingExpr,
    MappingField,
    SequenceExpr,
    SInt,
    SRational,
    SReal,
    StringExpr,
    TokenSpans,
)

_DEFAULT_POSITIONAL_FIRST_FIELD_NAMES = frozenset({"id", "name"})
_ATOM_RE = re.compile(r"^[^\s\(\)\[\]\{\}\",;]+$")


def dumps(
    value: Any,
    *,
    indent: int | None = None,
    max_line_length: int = 88,
    positional_first_id_or_name: bool = False,
    positional_first_field_names: Iterable[str] | None = None,
    positional_single_field: bool = True,
    preserve_spans: bool = True,
) -> str:
    """Serialize a value to Mu source text.

    Args:
        value: Value to serialize. Supports Mu AST nodes, dataclass instances,
            and common Python container/scalar types.
        indent: Pretty indentation width. If `None`, concise mode is used.
        max_line_length: Preferred maximum line width before multiline layout.
        positional_first_id_or_name: When `True`, dataclass groups may render
            the first field positionally when its field name matches
            `positional_first_field_names`.
        positional_first_field_names: Field names that may be rendered
            positionally for the first dataclass field when
            `positional_first_id_or_name=True`.
        positional_single_field: When `True`, single-field dataclasses are
            rendered positionally (for example `(include "path")`).
        preserve_spans: When `True`, AST inputs with complete span metadata are
            rendered from source spans to preserve original formatting.
    """
    if max_line_length < 20:
        raise ValueError("max_line_length must be >= 20")

    pretty = indent is not None
    indent_width = 2 if indent is None else indent
    if indent_width <= 0:
        raise ValueError("indent must be > 0 when provided")
    positional_names = (
        _DEFAULT_POSITIONAL_FIRST_FIELD_NAMES
        if positional_first_field_names is None
        else frozenset(positional_first_field_names)
    )

    settings = _SerializeSettings(
        positional_first_id_or_name=positional_first_id_or_name,
        positional_first_field_names=positional_names,
        positional_single_field=positional_single_field,
        preserve_spans=preserve_spans,
    )
    doc = _to_document(value, settings=settings)

    formatter = _Formatter(
        pretty=pretty,
        indent=indent_width,
        max_line_length=max_line_length,
        preserve_spans=preserve_spans,
    )
    return formatter.format_document(doc)


def dumps_pretty(
    value: Any,
    *,
    indent: int = 2,
    max_line_length: int = 88,
    positional_first_id_or_name: bool = False,
    positional_first_field_names: Iterable[str] | None = None,
    positional_single_field: bool = True,
    preserve_spans: bool = True,
) -> str:
    """Serialize a value using pretty multiline formatting."""
    return dumps(
        value,
        indent=indent,
        max_line_length=max_line_length,
        positional_first_id_or_name=positional_first_id_or_name,
        positional_first_field_names=positional_first_field_names,
        positional_single_field=positional_single_field,
        preserve_spans=preserve_spans,
    )


def dumps_concise(
    value: Any,
    *,
    max_line_length: int = 88,
    positional_first_id_or_name: bool = False,
    positional_first_field_names: Iterable[str] | None = None,
    positional_single_field: bool = True,
    preserve_spans: bool = True,
) -> str:
    """Serialize a value using concise formatting."""
    return dumps(
        value,
        indent=None,
        max_line_length=max_line_length,
        positional_first_id_or_name=positional_first_id_or_name,
        positional_first_field_names=positional_first_field_names,
        positional_single_field=positional_single_field,
        preserve_spans=preserve_spans,
    )


@dataclass(frozen=True)
class _SerializeSettings:
    positional_first_id_or_name: bool
    positional_first_field_names: frozenset[str]
    positional_single_field: bool
    preserve_spans: bool


@dataclass(frozen=True)
class _FieldSpec:
    field_name: str
    mu_name: str
    marker: type[Any] | None


def _to_document(value: Any, *, settings: _SerializeSettings) -> Document:
    if isinstance(value, Document):
        return value if settings.preserve_spans else value.drop_spans()

    if isinstance(value, Expr):
        expr = value if settings.preserve_spans else value.drop_spans()
        return Document([expr])

    if _is_top_level_expr_list(value):
        return Document([_encode_expr(v, context="field", settings=settings) for v in value])

    return Document([_encode_expr(value, context="field", settings=settings)])


def _is_top_level_expr_list(value: Any) -> bool:
    if not isinstance(value, (list, tuple)):
        return False
    if not value:
        return False
    return all(isinstance(item, Expr) or dataclasses.is_dataclass(item) for item in value)


def _encode_expr(value: Any, *, context: str, settings: _SerializeSettings) -> Expr:
    if isinstance(value, Expr):
        return value if settings.preserve_spans else value.drop_spans()

    if isinstance(value, Quoted):
        return _encode_expr(value.value, context=context, settings=settings)

    if dataclasses.is_dataclass(value):
        return _encode_dataclass(value, settings=settings)

    if value is None:
        return AtomExpr("none")
    if isinstance(value, bool):
        return AtomExpr("true" if value else "false")
    if isinstance(value, int):
        return SInt(value)
    if isinstance(value, float):
        return SReal(value)
    if isinstance(value, str):
        return _encode_string(value, context=context)
    if isinstance(value, (list, tuple)):
        return SequenceExpr(
            [
                _encode_expr(item, context="sequence", settings=settings)
                for item in value
            ]
        )
    if isinstance(value, dict):
        fields = []
        for key, item in value.items():
            key_expr = _encode_expr(key, context="map_key", settings=settings)
            value_expr = _encode_expr(item, context="map_value", settings=settings)
            fields.append(MappingField(key=key_expr, value=value_expr))
        return MappingExpr(fields)

    return AtomExpr(str(value))


def _encode_string(value: str, *, context: str) -> Expr:
    if context == "field":
        return StringExpr(value)
    if _is_atom_text(value):
        return AtomExpr(value)
    return StringExpr(value)


def _is_atom_text(value: str) -> bool:
    return bool(value) and _ATOM_RE.fullmatch(value) is not None


def _encode_dataclass(value: Any, *, settings: _SerializeSettings) -> GroupExpr:
    target = type(value)
    fields = _collect_dataclass_fields(target, value, settings=settings)
    values: list[Expr] = [AtomExpr(_dataclass_tag(target))]

    if not fields:
        return GroupExpr(values=values)

    positional = _positional_fields(fields, settings=settings)

    for spec, encoded_values in fields:
        if spec.field_name in positional:
            values.extend(encoded_values)
            continue
        values.append(AtomExpr(f":{spec.mu_name}"))
        values.extend(encoded_values)

    return GroupExpr(values=values)


def _collect_dataclass_fields(
    target: type[Any],
    value: Any,
    *,
    settings: _SerializeSettings,
) -> list[tuple[_FieldSpec, list[Expr]]]:
    specs = _field_specs(target)
    result: list[tuple[_FieldSpec, list[Expr]]] = []
    for spec in specs:
        raw = getattr(value, spec.field_name)

        if spec.marker is OptionalArg and raw is None:
            continue

        if spec.marker in {ZeroOrMore, OneOrMore}:
            if not isinstance(raw, (list, tuple)):
                raise TypeError(
                    f"Field '{spec.field_name}' uses a vararg marker but value is not a sequence"
                )
            if spec.marker is OneOrMore and len(raw) == 0:
                raise ValueError(f"Field '{spec.field_name}' uses OneOrMore but value is empty")
            if len(raw) == 0:
                continue
            encoded = [
                _encode_expr(item, context="field", settings=settings)
                for item in raw
            ]
            result.append((spec, encoded))
            continue

        encoded_value = _encode_expr(raw, context="field", settings=settings)
        result.append((spec, [encoded_value]))

    return result


def _positional_fields(
    fields: list[tuple[_FieldSpec, list[Expr]]],
    *,
    settings: _SerializeSettings,
) -> set[str]:
    if not fields:
        return set()
    if settings.positional_single_field and len(fields) == 1:
        return {fields[0][0].field_name}

    if settings.positional_first_id_or_name:
        first = fields[0][0]
        names = {first.field_name, first.mu_name}
        if names & settings.positional_first_field_names:
            return {first.field_name}
    return set()


def _field_specs(target: type[Any]) -> list[_FieldSpec]:
    type_hints = typing.get_type_hints(target, include_extras=True)
    specs: list[_FieldSpec] = []
    for field in dataclasses.fields(target):
        hint = type_hints.get(field.name, Any)
        _, metadata = _unwrap_annotated(hint)

        name_meta = _find_metadata(metadata, FieldName)
        mu_name = name_meta.name if name_meta is not None else _snake_to_kebab(field.name)
        marker = _pick_arity_marker(metadata)

        specs.append(
            _FieldSpec(
                field_name=field.name,
                mu_name=mu_name,
                marker=marker,
            )
        )
    return specs


def _unwrap_annotated(target: Any) -> tuple[Any, list[Any]]:
    metadata: list[Any] = []
    current = target
    while get_origin(current) is Annotated:
        args = list(get_args(current))
        if not args:
            break
        current = args[0]
        metadata.extend(args[1:])
    return current, metadata


def _find_metadata(metadata: list[Any], target_type: type[Any]) -> Any | None:
    found: Any | None = None
    for item in metadata:
        if isinstance(item, target_type):
            if found is not None:
                raise TypeError(f"Multiple {target_type.__name__} annotations are not allowed")
            found = item
    return found


def _pick_arity_marker(metadata: list[Any]) -> type[Any] | None:
    markers: list[type[Any]] = []
    if any(_is_marker(item, OptionalArg) for item in metadata):
        markers.append(OptionalArg)
    if any(_is_marker(item, ZeroOrMore) for item in metadata):
        markers.append(ZeroOrMore)
    if any(_is_marker(item, OneOrMore) for item in metadata):
        markers.append(OneOrMore)
    if len(markers) > 1:
        names = ", ".join(marker.__name__ for marker in markers)
        raise TypeError(f"At most one arity marker is allowed, found: {names}")
    return markers[0] if markers else None


def _is_marker(item: Any, marker: type[Any]) -> bool:
    return item is marker or isinstance(item, marker)


def _dataclass_tag(target: type[Any]) -> str:
    override = getattr(target, "__mu_tag__", None)
    if override is not None:
        return str(override)
    return _snake_to_kebab(_camel_to_snake(target.__name__))


def _camel_to_snake(name: str) -> str:
    step1 = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", step1).lower()


def _snake_to_kebab(name: str) -> str:
    return name.replace("_", "-")


class _Formatter:
    def __init__(
        self,
        *,
        pretty: bool,
        indent: int,
        max_line_length: int,
        preserve_spans: bool,
    ) -> None:
        self.pretty = pretty
        self.indent = indent
        self.max_line_length = max_line_length
        self.preserve_spans = preserve_spans

    def format_document(self, doc: Document) -> str:
        if self.preserve_spans:
            return self._format_document_preserving_spans(doc)
        parts = [self._format_expr_no_spans(expr, level=0) for expr in doc.exprs]
        if self.pretty:
            return "\n\n".join(parts)
        return "\n".join(parts)

    def _format_document_preserving_spans(self, doc: Document) -> str:
        result = ""
        if doc.leading_space is not None:
            result += doc.leading_space.raw

        for index, expr in enumerate(doc.exprs):
            result += self._render_expr_with_spans(expr, level=0)
            if index < len(doc.exprs) - 1 and not result.endswith((" ", "\n", "\t", "\r")):
                result += "\n\n" if self.pretty else "\n"
        return result

    def _render_expr_with_spans(self, expr: Expr, *, level: int) -> str:
        if isinstance(expr, (AtomExpr, StringExpr, SInt, SReal, SRational)):
            if _complete_token_spans(expr.span):
                return str(expr)
            return self._format_expr_no_spans(expr, level=level)

        if isinstance(expr, GroupExpr):
            return self._render_group_with_spans(expr, level=level)

        if isinstance(expr, SequenceExpr):
            return self._render_sequence_with_spans(expr, level=level)

        if isinstance(expr, MappingExpr):
            return self._render_map_with_spans(expr, level=level)

        return self._format_expr_no_spans(expr, level=level)

    def _render_group_with_spans(self, expr: GroupExpr, *, level: int) -> str:
        if expr.open_bracket is None and expr.close_bracket is None and expr.separators is None:
            return self._format_expr_no_spans(expr, level=level)

        open_text = str(expr.open_bracket) if _complete_token_spans(expr.open_bracket) else "("
        close_text = str(expr.close_bracket) if _complete_token_spans(expr.close_bracket) else ")"
        seps = expr.separators or []

        result = open_text
        for i, value in enumerate(expr.values):
            child_has_spans = _has_any_spans_expr(value)
            result += self._render_expr_with_spans(value, level=level + 1)
            sep = seps[i] if i < len(seps) else None
            if _complete_token_spans(sep):
                sep_text = str(sep)
                if sep_text == "" and i < len(expr.values) - 1 and not child_has_spans:
                    sep_text = " "
                result += sep_text
            elif i < len(expr.values) - 1:
                result += " "
        result += close_text
        return result

    def _render_sequence_with_spans(self, expr: SequenceExpr, *, level: int) -> str:
        if expr.open_bracket is None and expr.close_bracket is None and expr.separators is None:
            return self._format_expr_no_spans(expr, level=level)

        open_text = str(expr.open_bracket) if _complete_token_spans(expr.open_bracket) else "["
        close_text = str(expr.close_bracket) if _complete_token_spans(expr.close_bracket) else "]"
        seps = expr.separators or []

        result = open_text
        for i, value in enumerate(expr.values):
            child_has_spans = _has_any_spans_expr(value)
            result += self._render_expr_with_spans(value, level=level + 1)
            sep = seps[i] if i < len(seps) else None
            if _complete_token_spans(sep):
                sep_text = str(sep)
                if sep_text == "" and i < len(expr.values) - 1 and not child_has_spans:
                    sep_text = " "
                result += sep_text
            elif i < len(expr.values) - 1:
                result += " "
        result += close_text
        return result

    def _render_map_with_spans(self, expr: MappingExpr, *, level: int) -> str:
        if expr.open_bracket is None and expr.close_bracket is None and expr.separators is None:
            return self._format_expr_no_spans(expr, level=level)

        open_text = str(expr.open_bracket) if _complete_token_spans(expr.open_bracket) else "{"
        close_text = str(expr.close_bracket) if _complete_token_spans(expr.close_bracket) else "}"
        seps = expr.separators or []

        result = open_text
        for i, field in enumerate(expr.values):
            key_has_spans = _has_any_spans_expr(field.key)
            result += self._render_expr_with_spans(field.key, level=level + 1)
            if _complete_token_spans(field.separator):
                sep_text = str(field.separator)
                if sep_text == "" and not key_has_spans:
                    sep_text = ": "
                result += sep_text
            else:
                result += ": "
            result += self._render_expr_with_spans(field.value, level=level + 1)

            sep = seps[i] if i < len(seps) else None
            if _complete_token_spans(sep):
                sep_text = str(sep)
                if sep_text == "" and i < len(expr.values) - 1:
                    sep_text = ", "
                result += sep_text
            elif i < len(expr.values) - 1:
                result += ", "

        result += close_text
        return result

    def _format_expr_no_spans(self, expr: Expr, *, level: int) -> str:
        if self.pretty:
            return self._format_expr_pretty(expr, level=level)
        return self._format_expr_concise(expr, level=level)

    def _format_expr_concise(self, expr: Expr, *, level: int) -> str:
        inline = self._inline_expr(expr)
        if len(inline) <= self._budget(level):
            return inline
        return self._format_expr_pretty(expr, level=level)

    def _format_expr_pretty(self, expr: Expr, *, level: int) -> str:
        inline = self._inline_expr(expr)
        if len(inline) <= self._budget(level) and self._can_inline(expr):
            return inline

        if isinstance(expr, GroupExpr):
            return self._format_group(expr, level=level)
        if isinstance(expr, SequenceExpr):
            return self._format_sequence(expr, level=level)
        if isinstance(expr, MappingExpr):
            return self._format_map(expr, level=level)
        return inline

    def _format_group(self, expr: GroupExpr, *, level: int) -> str:
        if not expr.values:
            return "()"

        head = expr.values[0]
        args = expr.values[1:]
        inline = self._inline_expr(expr)
        if len(inline) <= self._budget(level):
            return inline

        parsed = self._parse_named_args(args)
        positional = parsed[0]
        named = parsed[1]

        head_text = self._inline_expr(head)
        start_tokens = [head_text] + [self._format_expr_concise(v, level=level) for v in positional]
        lines = ["(" + " ".join(start_tokens)]

        for name, values in named:
            field_lines = self._format_named_field(name, values, level=level + 1)
            lines.extend(self._indent_lines(field_lines, levels=1))

        lines.append(")")
        return "\n".join(lines)

    def _format_named_field(self, name: str, values: list[Expr], *, level: int) -> list[str]:
        if not values:
            return [f":{name}"]

        rendered = [self._format_expr_no_spans(v, level=level) for v in values]
        if all("\n" not in text for text in rendered):
            candidate = f":{name} " + " ".join(rendered)
            if len(candidate) <= self._budget(level):
                return [candidate]

        if len(rendered) == 1:
            lines = rendered[0].splitlines()
            first = f":{name} {lines[0]}"
            return [first, *lines[1:]]

        result = [f":{name}"]
        for text in rendered:
            result.extend(self._indent_lines(text.splitlines(), levels=1))
        return result

    def _format_sequence(self, expr: SequenceExpr, *, level: int) -> str:
        if not expr.values:
            return "[]"

        inline = self._inline_expr(expr)
        if len(inline) <= self._budget(level) and self._can_inline_sequence(expr):
            return inline

        lines = ["["]
        for value in expr.values:
            value_lines = self._format_expr_no_spans(value, level=level + 1).splitlines()
            lines.extend(self._indent_lines(value_lines, levels=1))
        lines.append("]")
        return "\n".join(lines)

    def _format_map(self, expr: MappingExpr, *, level: int) -> str:
        if not expr.values:
            return "{}"

        inline = self._inline_expr(expr)
        if len(inline) <= self._budget(level) and self._can_inline_map(expr):
            return inline

        lines = ["{"]
        for index, field in enumerate(expr.values):
            suffix = "," if index < len(expr.values) - 1 else ""
            key_text = self._format_expr_concise(field.key, level=level + 1)
            value_text = self._format_expr_no_spans(field.value, level=level + 1)
            value_lines = value_text.splitlines()
            if len(value_lines) == 1:
                lines.append(f"{self._indent(1)}{key_text}: {value_lines[0]}{suffix}")
                continue
            lines.append(f"{self._indent(1)}{key_text}: {value_lines[0]}")
            lines.extend(self._indent_lines(value_lines[1:], levels=1))
            if suffix:
                lines[-1] = f"{lines[-1]}{suffix}"
        lines.append("}")
        return "\n".join(lines)

    def _parse_named_args(self, values: list[Expr]) -> tuple[list[Expr], list[tuple[str, list[Expr]]]]:
        positional: list[Expr] = []
        named: list[tuple[str, list[Expr]]] = []
        current_name: str | None = None
        current_values: list[Expr] = []

        for value in values:
            if isinstance(value, AtomExpr) and value.value.startswith(":"):
                if current_name is not None:
                    named.append((current_name, current_values))
                current_name = value.value[1:]
                current_values = []
                continue

            if current_name is None:
                positional.append(value)
            else:
                current_values.append(value)

        if current_name is not None:
            named.append((current_name, current_values))

        return positional, named

    def _can_inline(self, expr: Expr) -> bool:
        if isinstance(expr, MappingExpr):
            return len(expr.values) <= 1
        if isinstance(expr, SequenceExpr):
            return self._can_inline_sequence(expr)
        return True

    def _can_inline_sequence(self, expr: SequenceExpr) -> bool:
        if not expr.values:
            return True
        if len(expr.values) > 8:
            return False
        if not all(isinstance(v, (AtomExpr, StringExpr, SInt, SReal, SRational)) for v in expr.values):
            return False
        first_type = type(expr.values[0])
        return all(type(v) is first_type for v in expr.values)

    def _can_inline_map(self, expr: MappingExpr) -> bool:
        if len(expr.values) <= 1:
            return True
        if not self.pretty:
            return True
        return False

    def _budget(self, level: int) -> int:
        return self.max_line_length - (self.indent * level)

    def _inline_expr(self, expr: Expr) -> str:
        if isinstance(expr, AtomExpr):
            return expr.value
        if isinstance(expr, SInt):
            return str(expr.value)
        if isinstance(expr, SReal):
            return repr(expr.value)
        if isinstance(expr, SRational):
            return f"{expr.value[0]}/{expr.value[1]}"
        if isinstance(expr, StringExpr):
            return _quote_string(expr.value)
        if isinstance(expr, GroupExpr):
            inner = " ".join(self._inline_expr(item) for item in expr.values)
            return f"({inner})"
        if isinstance(expr, SequenceExpr):
            inner = " ".join(self._inline_expr(item) for item in expr.values)
            return f"[{inner}]"
        if isinstance(expr, MappingExpr):
            parts = [
                f"{self._inline_expr(field.key)}: {self._inline_expr(field.value)}"
                for field in expr.values
            ]
            return "{" + ", ".join(parts) + "}"
        raise TypeError(f"Unsupported expression type: {type(expr).__name__}")

    def _indent(self, levels: int) -> str:
        return " " * (self.indent * levels)

    def _indent_lines(self, lines: list[str], *, levels: int) -> list[str]:
        prefix = self._indent(levels)
        return [f"{prefix}{line}" if line else prefix for line in lines]


def _quote_string(value: str) -> str:
    escaped = (
        value.replace("\\", "\\\\")
        .replace("\n", "\\n")
        .replace("\t", "\\t")
        .replace("\r", "\\r")
        .replace("\0", "\\0")
        .replace('"', '\\"')
    )
    return f'"{escaped}"'


def _complete_token_spans(spans: TokenSpans | None) -> bool:
    if spans is None:
        return False
    return spans.token is not None and spans.space is not None


def _has_any_spans_expr(expr: Expr) -> bool:
    if isinstance(expr, (AtomExpr, StringExpr, SInt, SReal, SRational)):
        return expr.span is not None
    if isinstance(expr, GroupExpr):
        return (
            expr.open_bracket is not None
            or expr.close_bracket is not None
            or expr.separators is not None
            or any(_has_any_spans_expr(value) for value in expr.values)
        )
    if isinstance(expr, SequenceExpr):
        return (
            expr.open_bracket is not None
            or expr.close_bracket is not None
            or expr.separators is not None
            or any(_has_any_spans_expr(value) for value in expr.values)
        )
    if isinstance(expr, MappingExpr):
        return (
            expr.open_bracket is not None
            or expr.close_bracket is not None
            or expr.separators is not None
            or any(
                field.separator is not None
                or _has_any_spans_expr(field.key)
                or _has_any_spans_expr(field.value)
                for field in expr.values
            )
        )
    return False


__all__ = ["dumps", "dumps_concise", "dumps_pretty"]



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import ClassVar, Generic, TypeVar

ArgumentName = TypeVar("ArgumentName")
ArgumentValue = TypeVar("ArgumentValue")


@dataclass(frozen=True)
class ArgArity:
    min_required: int
    max_required: int | None
    Required: ClassVar[ArgArity]
    Optional: ClassVar[ArgArity]
    ZeroOrMore: ClassVar[ArgArity]
    OneOrMore: ClassVar[ArgArity]

    def __post_init__(self) -> None:
        if self.min_required < 0:
            raise ValueError("min_required must be non-negative")
        if self.max_required is not None and self.max_required < self.min_required:
            raise ValueError("max_required must be >= min_required")

    @property
    def is_vararg(self) -> bool:
        return self.max_required is None

    @property
    def is_nullable(self) -> bool:
        return self.min_required == 0

    def consume(self, consumed: int) -> ArgArity:
        if consumed < 0:
            raise ValueError(f"consumed must be non-negative: {consumed}")
        if consumed == 0:
            return self
        new_min_required = max(0, self.min_required - consumed)
        if self.max_required is None:
            new_max_required: int | None = None
        else:
            if consumed > self.max_required:
                raise ValueError(
                    f"consumed must be <= max_required: {consumed} > {self.max_required}"
                )
            new_max_required = self.max_required - consumed
        return ArgArity(new_min_required, new_max_required)

    def as_suffix_string(self) -> str:
        if self.min_required == 1 and self.max_required == 1:
            return ""
        if self.min_required == 0 and self.max_required == 1:
            return "?"
        if self.min_required == 1 and self.max_required is None:
            return "+"
        if self.min_required == 0 and self.max_required is None:
            return "*"
        if self.max_required is None:
            return f"{{{self.min_required},}}"
        return f"{{{self.min_required}..{self.max_required}}}"


ArgArity.Required = ArgArity(1, 1)  # type: ignore[attr-defined]
ArgArity.Optional = ArgArity(0, 1)  # type: ignore[attr-defined]
ArgArity.ZeroOrMore = ArgArity(0, None)  # type: ignore[attr-defined]
ArgArity.OneOrMore = ArgArity(1, None)  # type: ignore[attr-defined]


class MatchArgsException(Exception):
    pass


class TooManyArgumentsException(MatchArgsException):
    def __init__(self, values: Sequence[object]):
        self.values = list(values)
        super().__init__(
            f"Too many arguments provided. Extra: {', '.join(map(str, self.values))}"
        )


class MissingArgumentException(MatchArgsException):
    def __init__(self, names: Sequence[object]):
        self.names = list(names)
        super().__init__(f"Missing required argument: '{self.names}'")


class DuplicateArgumentException(MatchArgsException):
    def __init__(self, name: object):
        self.name = name
        super().__init__(f"Duplicate argument provided for: '{name}'")


class PositionalArgumentsAfterOutOfOrderNamedArgumentException(MatchArgsException):
    def __init__(self, arg_value: str, arg_index: int, named_arg_index: int):
        self.arg_value = arg_value
        self.arg_index = arg_index
        self.named_arg_index = named_arg_index
        super().__init__(
            "Positional argument "
            f"'{arg_value}' (at index {arg_index}) cannot follow an out-of-order named "
            f"argument (last one at index {named_arg_index})."
        )


class UnknownArgumentException(MatchArgsException):
    def __init__(self, names: Sequence[object]):
        self.names = list(names)
        super().__init__(f"Unknown named argument: {self.names}")


class NamedArgumentRequiresValueException(MatchArgsException):
    def __init__(self, names: Sequence[object]):
        self.names = list(names)
        super().__init__(f"Named argument {self.names} requires at least one value.")


class AmbiguousPositionalArgumentException(MatchArgsException):
    def __init__(
        self,
        arg_value: str | None = None,
        possible_params: list[str] | None = None,
        details: str = "Ambiguous assignment for positional arguments.",
    ):
        self.arg_value = arg_value
        self.possible_params = possible_params
        message = details
        if arg_value is not None:
            message += f" Argument: '{arg_value}'."
        if possible_params is not None:
            message += f" Could belong to: {', '.join(possible_params)}."
        super().__init__(message)


class NamedArgumentValueExpectedException(MatchArgsException):
    def __init__(self, name: str, unexpected_arg: str):
        self.name = name
        self.unexpected_arg = unexpected_arg
        super().__init__(
            f"Named argument ':{name}' expected a value, but got named argument ':{unexpected_arg}'."
        )


@dataclass(frozen=True)
class NamedArg(Generic[ArgumentName]):
    name: ArgumentName


@dataclass(frozen=True)
class PositionalArg(Generic[ArgumentValue]):
    value: ArgumentValue


class GoResult:
    @property
    def severity(self) -> int:
        raise NotImplementedError


@dataclass(frozen=True)
class GoSuccess(GoResult, Generic[ArgumentName, ArgumentValue]):
    result: dict[ArgumentName, list[ArgumentValue]]

    @property
    def severity(self) -> int:
        return 0


@dataclass(frozen=True)
class GoMissing(GoResult, Generic[ArgumentName, ArgumentValue]):
    missing: list[tuple[ArgArity, ArgumentName]]

    @property
    def severity(self) -> int:
        return 2


@dataclass(frozen=True)
class GoTooMany(GoResult, Generic[ArgumentName, ArgumentValue]):
    arg_name: ArgumentName
    extra_values: list[ArgumentValue]
    is_duplicate: bool

    @property
    def severity(self) -> int:
        return 1


@dataclass(frozen=True)
class GoPositionalInNamedMode(GoResult, Generic[ArgumentName, ArgumentValue]):
    arg_value: ArgumentValue
    due_to_named_out_of_order: bool

    @property
    def severity(self) -> int:
        return 5


@dataclass(frozen=True)
class GoUnknownPositionalArgument(GoResult, Generic[ArgumentName, ArgumentValue]):
    arg_value: ArgumentValue

    @property
    def severity(self) -> int:
        return 5


@dataclass(frozen=True)
class GoAmbiguous(GoResult, Generic[ArgumentName, ArgumentValue]):
    arg_name: ArgumentName

    @property
    def severity(self) -> int:
        return 10


class ParameterPosition:
    def inc_if_positional(self) -> ParameterPosition:
        if isinstance(self, PositionalPosition):
            return PositionalPosition(self.parameter_index + 1)
        return self


@dataclass(frozen=True)
class NamedPosition(ParameterPosition):
    due_to_named_out_of_order: bool


@dataclass(frozen=True)
class PositionalPosition(ParameterPosition):
    parameter_index: int


@dataclass(frozen=True)
class ConsumptionMode:
    min: int
    greedy: bool


def _result_map_put(
    result: dict[ArgumentName, list[ArgumentValue]],
    key: ArgumentName,
    values: list[ArgumentValue],
) -> dict[ArgumentName, list[ArgumentValue]]:
    updated = {k: list(v) for k, v in result.items()}
    updated[key] = values
    return updated


def match_args(
    parameters: list[tuple[ArgArity, ArgumentName]],
    arguments: list[NamedArg[ArgumentName] | PositionalArg[ArgumentValue]],
) -> dict[ArgumentName, list[ArgumentValue]]:
    parameter_names = [name for _, name in parameters]
    if len(set(parameter_names)) != len(parameters):
        raise ValueError(f"Duplicate parameter names found: {parameter_names}")

    param_to_arity: dict[ArgumentName, ArgArity] = {name: arity for arity, name in parameters}

    invalid_named_args = [
        arg.name
        for arg in arguments
        if isinstance(arg, NamedArg) and arg.name not in param_to_arity
    ]
    if invalid_named_args:
        raise UnknownArgumentException(invalid_named_args)

    empty_named_args: list[object] = []
    for index, arg in enumerate(arguments):
        if not isinstance(arg, NamedArg):
            continue
        if index + 1 >= len(arguments) or isinstance(arguments[index + 1], NamedArg):
            empty_named_args.append(arg.name)
    if empty_named_args:
        raise NamedArgumentRequiresValueException(empty_named_args)

    def go(
        depth: int,
        position: ParameterPosition,
        argument_index: int,
        result: dict[ArgumentName, list[ArgumentValue]],
    ) -> GoResult:
        if depth > len(parameters) + len(arguments):
            raise RuntimeError(
                f"Recursion depth exceeded: {depth} > {len(parameters) + len(arguments)}"
            )

        def consume(
            arg_position: ParameterPosition,
            arg_param_name: ArgumentName,
            mode: ConsumptionMode,
            skip_args: int,
            new_values: list[ArgumentValue],
        ) -> GoResult:
            current_value_list = result[arg_param_name]
            param_arity = param_to_arity[arg_param_name]
            effective_arity = param_arity.consume(len(current_value_list))
            new_value_count = len(new_values)

            if effective_arity.max_required == 0:
                return GoTooMany(
                    arg_param_name,
                    new_values,
                    param_arity.max_required == 1,
                )

            min_consumed = max(mode.min, min(effective_arity.min_required, new_value_count))
            if effective_arity.max_required is not None:
                max_consumable = min(effective_arity.max_required, new_value_count)
            else:
                max_consumable = new_value_count

            if mode.greedy:
                min_consumed = max(min_consumed, max_consumable)

            best: GoResult | None = None

            def consider(case: GoResult) -> GoAmbiguous | None:
                nonlocal best
                if isinstance(case, GoAmbiguous):
                    return case

                if not isinstance(best, GoSuccess):
                    if isinstance(case, GoSuccess):
                        best = case
                    else:
                        if best is None:
                            best = case
                        elif best.severity < case.severity:
                            best = case
                else:
                    if isinstance(case, GoSuccess) and case != best:
                        return GoAmbiguous(arg_param_name)
                return None

            for consumed in range(min_consumed, max_consumable + 1):
                final_arity = effective_arity.consume(consumed)
                new_result = _result_map_put(
                    result,
                    arg_param_name,
                    current_value_list + new_values[:consumed],
                )

                case = go(
                    depth + 1,
                    arg_position.inc_if_positional(),
                    argument_index + skip_args + consumed,
                    new_result,
                )
                r = consider(case)
                if r is not None:
                    return r

                if final_arity.max_required != 0 and consumed > 0:
                    case2 = go(
                        depth + 1,
                        arg_position,
                        argument_index + skip_args + consumed,
                        new_result,
                    )
                    r2 = consider(case2)
                    if r2 is not None:
                        return r2

            if best is None:
                raise RuntimeError("Expected at least one branch result")
            return best

        if argument_index >= len(arguments):
            missing_params = [
                (arity, name)
                for arity, name in parameters
                if arity.consume(len(result[name])).min_required > 0
            ]
            if missing_params:
                return GoMissing(missing_params)
            return GoSuccess(result)

        if isinstance(position, PositionalPosition) and position.parameter_index >= len(parameters):
            return go(depth + 1, NamedPosition(due_to_named_out_of_order=False), argument_index, result)

        arg = arguments[argument_index]
        if isinstance(arg, PositionalArg):
            if isinstance(position, PositionalPosition):
                _, param_name = parameters[position.parameter_index]
                return consume(
                    arg_position=position,
                    arg_param_name=param_name,
                    mode=ConsumptionMode(min=0, greedy=False),
                    skip_args=0,
                    new_values=[arg.value],
                )
            assert isinstance(position, NamedPosition)
            return GoPositionalInNamedMode(arg.value, position.due_to_named_out_of_order)

        arg_param_name = arg.name
        new_values: list[ArgumentValue] = []
        for next_arg in arguments[argument_index + 1 :]:
            if isinstance(next_arg, NamedArg):
                break
            new_values.append(next_arg.value)

        greedy = True

        if isinstance(position, NamedPosition):
            return consume(
                arg_position=position,
                arg_param_name=arg_param_name,
                mode=ConsumptionMode(min=1, greedy=greedy),
                skip_args=1,
                new_values=new_values,
            )

        assert isinstance(position, PositionalPosition)
        _, param_name = parameters[position.parameter_index]
        if arg_param_name == param_name:
            return consume(
                arg_position=position,
                arg_param_name=arg_param_name,
                mode=ConsumptionMode(min=1, greedy=greedy),
                skip_args=1,
                new_values=new_values,
            )

        new_parameter_index = next(
            (i for i, (_, name) in enumerate(parameters) if name == arg_param_name),
            -1,
        )
        if new_parameter_index == -1:
            raise RuntimeError("Invariant violated: named arg should have been validated")
        if new_parameter_index == position.parameter_index:
            raise RuntimeError("Invariant violated: expected mismatched parameter index")

        if new_parameter_index < position.parameter_index:
            return consume(
                arg_position=NamedPosition(due_to_named_out_of_order=True),
                arg_param_name=arg_param_name,
                mode=ConsumptionMode(min=1, greedy=greedy),
                skip_args=1,
                new_values=new_values,
            )

        all_optional = all(
            parameters[i][0].is_nullable
            for i in range(position.parameter_index, new_parameter_index)
        )
        if all_optional:
            return consume(
                arg_position=PositionalPosition(new_parameter_index),
                arg_param_name=arg_param_name,
                mode=ConsumptionMode(min=1, greedy=greedy),
                skip_args=1,
                new_values=new_values,
            )

        return consume(
            arg_position=NamedPosition(due_to_named_out_of_order=True),
            arg_param_name=arg_param_name,
            mode=ConsumptionMode(min=1, greedy=greedy),
            skip_args=1,
            new_values=new_values,
        )

    initial_map: dict[ArgumentName, list[ArgumentValue]] = {
        name: [] for _, name in parameters
    }
    result = go(
        depth=0,
        position=PositionalPosition(0),
        argument_index=0,
        result=initial_map,
    )

    if isinstance(result, GoSuccess):
        return result.result
    if isinstance(result, GoMissing):
        raise MissingArgumentException([str(name) for _, name in result.missing])
    if isinstance(result, GoUnknownPositionalArgument):
        raise UnknownArgumentException([str(result.arg_value)])
    if isinstance(result, GoTooMany):
        if result.is_duplicate:
            raise DuplicateArgumentException(str(result.arg_name))
        raise TooManyArgumentsException([str(v) for v in result.extra_values])
    if isinstance(result, GoAmbiguous):
        raise AmbiguousPositionalArgumentException(str(result.arg_name))
    if isinstance(result, GoPositionalInNamedMode):
        if result.due_to_named_out_of_order:
            raise PositionalArgumentsAfterOutOfOrderNamedArgumentException(
                str(result.arg_value),
                0,
                0,
            )
        raise TooManyArgumentsException([str(result.arg_value)])

    raise RuntimeError(f"Unexpected go result: {result!r}")



"""Stable public API exports for the `mu` package."""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

from mu.loading import load as load
from mu.loading import loads as loads
from mu.parser import ParseError as ParseError
from mu.parser import parse as parse
from mu.printer import dumps as dumps
from mu.printer import dumps_concise as dumps_concise
from mu.printer import dumps_pretty as dumps_pretty
from mu.quoted import Quoted as Quoted
from mu.typed import DecodeContext as DecodeContext
from mu.typed import DecodeError as DecodeError
from mu.typed import DecoderFn as DecoderFn
from mu.typed import DecoderRegistry as DecoderRegistry
from mu.typed import DecodeWith as DecodeWith
from mu.typed import FieldName as FieldName
from mu.typed import OneOrMore as OneOrMore
from mu.typed import OptionalArg as OptionalArg
from mu.typed import ZeroOrMore as ZeroOrMore
from mu.typed import decode as decode
from mu.typed import parse_many as parse_many
from mu.typed import parse_one as parse_one
from mu.typed import tag as tag
from mu.types import AtomExpr as AtomExpr
from mu.types import Document as Document
from mu.types import Expr as Expr
from mu.types import GroupExpr as GroupExpr
from mu.types import MappingExpr as MappingExpr
from mu.types import MappingField as MappingField
from mu.types import SequenceExpr as SequenceExpr
from mu.types import StringExpr as StringExpr

try:
    __version__ = version("lang-mu")
except PackageNotFoundError:  # pragma: no cover - local editable installs
    __version__ = "0.3.2"

__all__ = [
    "ParseError",
    "Quoted",
    "DecodeContext",
    "DecodeError",
    "DecodeWith",
    "DecoderFn",
    "DecoderRegistry",
    "FieldName",
    "OneOrMore",
    "OptionalArg",
    "ZeroOrMore",
    "AtomExpr",
    "Document",
    "Expr",
    "GroupExpr",
    "MappingExpr",
    "MappingField",
    "SequenceExpr",
    "StringExpr",
    "__version__",
    "load",
    "loads",
    "dumps",
    "dumps_concise",
    "dumps_pretty",
    "decode",
    "tag",
    "parse_many",
    "parse_one",
    "parse",
]



from __future__ import annotations

from typing import Generic, TypeVar

T = TypeVar("T")


class Quoted(Generic[T]):
    """Wraps a value that should be passed through without runtime evaluation."""

    def __init__(self, value: T):
        self.value = value

    def __repr__(self) -> str:
        return f"Quoted({self.value!r})"




from dataclasses import dataclass


@dataclass
class Pos:
    line: int
    col: int
    index: int

    def forward_via(self, char: str):
        assert len(char) == 1

        if char == "\n":
            return Pos(self.line + 1, 1, self.index + 1)
        else:
            return Pos(self.line, self.col + 1, self.index + 1)

    def backward_via(self, char: str):
        assert len(char) == 1

        if char == "\n":
            raise ValueError("Cannot move backwards over a newline")
        else:
            return Pos(self.line, self.col - 1, self.index - 1)


@dataclass
class Span:
    start: Pos
    end: Pos
    raw: str

    def __post_init__(self):
        assert len(self.raw) == self.end.index - self.start.index

    def __getitem__(self, key):
        assert isinstance(key, slice)

        # 1. Extract raw string length
        length = len(self.raw)

        # 2. Convert slice.start and slice.stop to valid integer offsets
        #    handling None and negative values (Python-style).
        start_offset = key.start if key.start is not None else 0
        stop_offset = key.stop if key.stop is not None else length

        if start_offset < 0:
            start_offset = length + start_offset
        if stop_offset < 0:
            stop_offset = length + stop_offset

        # 3. Clamp to avoid out-of-range issues
        if start_offset < 0:
            start_offset = 0
        if start_offset > length:
            start_offset = length
        if stop_offset < 0:
            stop_offset = 0
        if stop_offset > length:
            stop_offset = length

        # 4. Prepare new start/end
        new_start = self.start
        new_end = self.end

        # Move new_start forward by start_offset
        for i in range(start_offset):
            new_start = new_start.forward_via(self.raw[i])

        # Move new_end backward for the part after stop_offset
        chars_to_trim_from_end = length - stop_offset
        for i in range(chars_to_trim_from_end):
            new_end = new_end.backward_via(self.raw[-i - 1])

        new_raw = self.raw[start_offset:stop_offset]
        return Span(new_start, new_end, new_raw)

    def __add__(self, other: "Span") -> "Span":
        """
        Merge two spans if they are contiguous, meaning:
        self.end.index == other.start.index.

        If not contiguous, raises ValueError.
        """
        if self.end.index != other.start.index:
            raise ValueError(
                f"Cannot add spans that are not contiguous: "
                f"self.end.index={self.end.index}, other.start.index={other.start.index}"
            )

        # Construct the new raw text.
        new_start = self.start
        new_raw = self.raw + other.raw
        new_end = other.end
        return Span(new_start, new_end, new_raw)


class _Input:
    EOS = "\0"

    def __init__(self, str: str):
        self.str = str

        self.index = 0
        self.line = 1
        self.column = 1

        self.mark_index = 0
        self.mark_line = 1
        self.mark_column = 1

        self.current = _Input.EOS if len(str) == 0 else str[0]

    @property
    def position(self) -> Pos:
        return Pos(self.line, self.column, self.index)

    def next(self):
        if self.index < len(self.str):
            if self.current == "\n":
                self.line += 1
                self.column = 1
            else:
                self.column += 1

            self.index += 1
            self.current = (
                _Input.EOS if self.index >= len(self.str) else self.str[self.index]
            )
        else:
            self.current = _Input.EOS

    def capture(self) -> Span:
        start_pos = Pos(self.mark_line, self.mark_column, self.mark_index)
        result = Span(start_pos, self.position, self.str[self.mark_index : self.index])
        self.mark_index = self.index
        self.mark_line = self.line
        self.mark_column = self.column
        return result

    def __repr__(self) -> str:
        start_index = max(0, self.index - 8)
        end_index = min(len(self.str), self.index + 8)

        chars = []
        for i in range(start_index, end_index):
            char = self.str[i]
            if char == " ":
                char = "·"
            if char == "\n":
                char = "\\n"
            elif char == "\t":
                char = "\\t"
            elif not char.isprintable():
                char = f"\\x{ord(char):02x}"

            if i == self.index:
                chars.append(f"[{char}]")
            else:
                chars.append(char)

        snippet = " ".join(chars)

        return f"Input(line={self.line}, column={self.column}, {snippet})"


DEBUG_ENABLED = False
DEBUG_DEPTH = 0


# debug decorator
def debug(func):
    if not DEBUG_ENABLED:
        return func

    def wrapper(*args, **kwargs):
        global DEBUG_DEPTH
        saved_depth = DEBUG_DEPTH
        prefix = "  " * DEBUG_DEPTH
        print(
            f"{prefix}{func.__name__}({', '.join(repr(x) for x in args)}, {kwargs}) {{"
        )
        try:
            DEBUG_DEPTH += 1
            result = func(*args, **kwargs)
            return result
        except Exception as e:
            exception = e
            raise e
        finally:
            if "exception" in locals():
                print(f"{prefix}}}raise {exception}")
            else:
                print(f"{'  ' * DEBUG_DEPTH}return {repr(result)}")
            DEBUG_DEPTH -= 1
            assert saved_depth == DEBUG_DEPTH
            print(f"{prefix}}}")

    return wrapper



from dataclasses import dataclass

from mu.input import Span

# from decimal import Decimal
# from fractions import Fraction


def _quote_string(value: str) -> str:
    escaped = (
        value.replace("\\", "\\\\")
        .replace("\n", "\\n")
        .replace("\t", "\\t")
        .replace("\r", "\\r")
        .replace("\0", "\\0")
        .replace('"', '\\"')
    )
    return f'"{escaped}"'


def _has_spans_expr(expr: "Expr") -> bool:
    if isinstance(expr, (AtomExpr, StringExpr, SReal, SInt, SRational)):
        return expr.span is not None
    if isinstance(expr, GroupExpr):
        return (
            expr.open_bracket is not None
            or expr.separators is not None
            or expr.close_bracket is not None
            or any(_has_spans_expr(value) for value in expr.values)
        )
    if isinstance(expr, SequenceExpr):
        return (
            expr.open_bracket is not None
            or expr.separators is not None
            or expr.close_bracket is not None
            or any(_has_spans_expr(value) for value in expr.values)
        )
    if isinstance(expr, MappingExpr):
        return (
            expr.open_bracket is not None
            or expr.separators is not None
            or expr.close_bracket is not None
            or any(
                field.separator is not None
                or _has_spans_expr(field.key)
                or _has_spans_expr(field.value)
                for field in expr.values
            )
        )
    return False


class Expr:
    """Base class for all Mu expression AST nodes."""

    Atom: type["AtomExpr"] = None  # type: ignore
    Str: type["StringExpr"] = None  # type: ignore
    Real: type["SReal"] = None  # type: ignore
    Int: type["SInt"] = None  # type: ignore
    Rational: type["SRational"] = None  # type: ignore

    Group: type["GroupExpr"] = None  # type: ignore
    Seq: type["SequenceExpr"] = None  # type: ignore
    Map: type["MappingExpr"] = None  # type: ignore

    def drop_spans(self) -> "Expr":
        """Return a structurally equivalent node with token/space spans removed."""
        match self:
            case AtomExpr(value=value):
                return AtomExpr(value)
            case StringExpr(value=value):
                return StringExpr(value)
            case SReal(value=value):
                return SReal(value)
            case SInt(value=value):
                return SInt(value)
            case SRational(value=value):
                return SRational(value)
            case GroupExpr(values=values):
                return GroupExpr([v.drop_spans() for v in values])
            case SequenceExpr(values=values):
                return SequenceExpr([v.drop_spans() for v in values])
            case MappingExpr(values=values):
                return MappingExpr(
                    [
                        MappingField(field.key.drop_spans(), field.value.drop_spans())
                        for field in values
                    ]
                )
        raise TypeError(f"Unsupported expression type: {type(self).__name__}")


@dataclass
class TokenSpans:
    """Captured token and trailing space spans for source-preserving parsing."""

    token: Span | None = None
    space: Span | None = None

    def __str__(self) -> str:
        result = ""
        if self.token is not None:
            result += f"{self.token.raw}"
        if self.space is not None:
            result += f"{self.space.raw}"
        return result


# a | ab | define-test | 123 | 123.456 | 123. | .456
@dataclass
class AtomExpr(Expr):
    """Atom expression (identifier-like token) node."""

    value: str
    span: TokenSpans | None = None

    def __str__(self) -> str:
        if self.span is not None:
            return str(self.span)
        return self.value


# "a" | "ab" | "define-test" | "123" | "123.456" | "123." | ".456"
# #"\d+" | #"\d+.\*" ... (raw)
@dataclass
class StringExpr(Expr):
    """String expression node (normal or raw Mu string literal)."""

    value: str
    span: TokenSpans | None = None

    def __str__(self) -> str:
        if self.span is not None:
            return str(self.span)
        return _quote_string(self.value)


# 123.456 | 123. | 123e4 | 123.456e4 | 123.e4
@dataclass
class SReal(Expr):
    """Floating-point token node."""

    value: float
    span: TokenSpans | None = None

    def __str__(self) -> str:
        if self.span is not None:
            return str(self.span)
        return str(self.value)


# 123 | 0x123 | 0b101 | 0o123
@dataclass
class SInt(Expr):
    """Integer token node."""

    value: int
    span: TokenSpans | None = None

    def __str__(self) -> str:
        if self.span is not None:
            return str(self.span)
        return str(self.value)


# 123/456
@dataclass
class SRational(Expr):
    """Rational token node represented as `(numerator, denominator)`."""

    value: tuple[int, int]
    span: TokenSpans | None = None

    def __str__(self) -> str:
        if self.span is not None:
            return str(self.span)
        return f"{self.value[0]}/{self.value[1]}"


# (a b c)
@dataclass
class GroupExpr(Expr):
    """Parenthesized expression group node."""

    values: list[Expr]

    open_bracket: TokenSpans | None = None
    separators: list[TokenSpans] | None = None
    close_bracket: TokenSpans | None = None

    def __str__(self) -> str:
        if (
            self.open_bracket is None
            and self.separators is None
            and self.close_bracket is None
        ):
            return "(" + " ".join(str(value) for value in self.values) + ")"

        result = ""
        result += "(" if self.open_bracket is None else f"{self.open_bracket}"

        separators = self.separators or []
        for i, value in enumerate(self.values):
            result += str(value)
            if i < len(separators):
                result += f"{separators[i]}"
            elif i < len(self.values) - 1:
                result += " "

        result += ")" if self.close_bracket is None else f"{self.close_bracket}"
        return result


# [a b c] | [a, b, c] | []
@dataclass
class SequenceExpr(Expr):
    """Bracketed sequence node."""

    values: list[Expr]

    open_bracket: TokenSpans | None = None
    separators: list[TokenSpans] | None = None
    close_bracket: TokenSpans | None = None

    def __str__(self) -> str:
        if (
            self.open_bracket is None
            and self.separators is None
            and self.close_bracket is None
        ):
            return "[" + " ".join(str(value) for value in self.values) + "]"

        result = ""
        result += "[" if self.open_bracket is None else f"{self.open_bracket}"

        separators = self.separators or []
        for i, value in enumerate(self.values):
            result += str(value)
            if i < len(separators):
                result += f"{separators[i]}"
            elif i < len(self.values) - 1:
                result += " "

        result += "]" if self.close_bracket is None else f"{self.close_bracket}"
        return result


@dataclass
class MappingField:
    """Single key/value field inside a mapping expression."""

    key: Expr
    value: Expr
    separator: TokenSpans | None = None

    def __str__(self) -> str:
        if self.separator is None:
            return f"{self.key}: {self.value}"
        return f"{self.key}{self.separator}{self.value}"


# { a : b, c : d }
@dataclass
class MappingExpr(Expr):
    """Brace-delimited mapping node."""

    values: list[MappingField]

    open_bracket: TokenSpans | None = None
    separators: list[TokenSpans] | None = None
    close_bracket: TokenSpans | None = None

    def __str__(self) -> str:
        if (
            self.open_bracket is None
            and self.separators is None
            and self.close_bracket is None
        ):
            return "{" + ", ".join(str(field) for field in self.values) + "}"

        result = ""
        result += "{" if self.open_bracket is None else f"{self.open_bracket}"

        separators = self.separators or []
        for i, field in enumerate(self.values):
            result += str(field)
            if i < len(separators):
                result += f"{separators[i]}"
            elif i < len(self.values) - 1:
                result += ", "

        result += "}" if self.close_bracket is None else f"{self.close_bracket}"
        return result


@dataclass
class Document:
    """Top-level Mu parse result containing one or more expressions."""

    exprs: list[Expr]
    leading_space: Span | None = None

    def __str__(self) -> str:
        if self.leading_space is None and not any(
            _has_spans_expr(expr) for expr in self.exprs
        ):
            return "\n".join(str(expr) for expr in self.exprs)

        result = ""
        if self.leading_space is not None:
            result += f"{self.leading_space.raw}"
        result += "".join(str(expr) for expr in self.exprs)
        return result

    def drop_spans(self) -> "Document":
        """Return an equivalent document with spans removed from all expressions."""
        return Document([expr.drop_spans() for expr in self.exprs], leading_space=None)


Expr.Atom = AtomExpr
Expr.Str = StringExpr
Expr.Real = SReal
Expr.Int = SInt
Expr.Rational = SRational

Expr.Group = GroupExpr
Expr.Seq = SequenceExpr
Expr.Map = MappingExpr



"""Typed decoding helpers that map Mu AST expressions to Python types."""

from __future__ import annotations

import dataclasses
import re
import types
import typing
from collections.abc import Callable
from dataclasses import MISSING, dataclass
from typing import Annotated, Any, TypeVar, get_args, get_origin

from mu.arg_match import (
    ArgArity,
    MatchArgsException,
    NamedArg,
    PositionalArg,
    match_args,
)
from mu.parser import parse
from mu.quoted import Quoted
from mu.types import (
    AtomExpr,
    Document,
    Expr,
    GroupExpr,
    MappingExpr,
    SequenceExpr,
    SInt,
    SRational,
    SReal,
    StringExpr,
)

T = TypeVar("T")
DecoderFn = Callable[[Expr, "DecodeContext"], Any]


class DecodeError(ValueError):
    """Structured decode error with path/expected/got/span/cause details."""

    def __init__(
        self,
        path: str,
        expected: str,
        got: str,
        span: Any | None = None,
        cause: Exception | None = None,
        message: str | None = None,
    ):
        self.path = path
        self.expected = expected
        self.got = got
        self.span = span
        self.cause = cause
        if message is None:
            message = f"{path}: expected {expected}, got {got}"
            if cause is not None:
                message += f" ({cause})"
        super().__init__(message)


@dataclass(frozen=True)
class FieldName:
    """Annotated marker for overriding a dataclass field's Mu argument name."""

    name: str


@dataclass(frozen=True)
class DecodeWith:
    """Annotated marker for a field-specific custom decode function."""

    fn: DecoderFn


class OptionalArg:
    """Annotated marker for optional field arity (`0..1`)."""

    pass


class ZeroOrMore:
    """Annotated marker for variadic field arity (`0..N`)."""

    pass


class OneOrMore:
    """Annotated marker for required variadic field arity (`1..N`)."""

    pass


@dataclass(frozen=True)
class DecodeContext:
    """Context object passed to custom decoder functions."""

    path: str
    target: Any
    registry: DecoderRegistry

    def decode(self, expr: Expr, target: Any, *, path: str | None = None) -> Any:
        """Decode a nested expression with inherited registry settings."""
        next_path = self.path if path is None else path
        return _decode_value(expr, target, self.registry, next_path)


class DecoderRegistry:
    """Registry of target-type-specific decoder functions."""

    def __init__(self) -> None:
        self._registry: dict[Any, DecoderFn] = {}

    def register(self, target: Any, fn: DecoderFn) -> None:
        """Register a custom decoder function for a target type."""
        self._registry[target] = fn

    def get(self, target: Any) -> DecoderFn | None:
        """Look up a custom decoder function for a target type."""
        return self._registry.get(target)


def tag(tag: str) -> Callable[[type[T]], type[T]]:
    """Decorator that overrides the default dataclass tag used during decoding."""

    def decorator(cls: type[T]) -> type[T]:
        typing.cast(Any, cls).__mu_tag__ = tag
        return cls

    return decorator


@dataclass(frozen=True)
class _FieldSpec:
    field_name: str
    mu_name: str
    type_hint: Any
    base_type: Any
    metadata: tuple[Any, ...]
    arity: ArgArity
    has_default: bool


def parse_one(
    source: str,
    target: Any,
    *,
    registry: DecoderRegistry | None = None,
) -> Any:
    """Parse and decode exactly one top-level Mu expression."""
    decode_registry = registry or DecoderRegistry()
    doc = parse(source, preserve_spans=True)
    if len(doc.exprs) != 1:
        raise DecodeError(
            path="$",
            expected="exactly one top-level expression",
            got=f"{len(doc.exprs)} expressions",
        )
    return _decode_value(doc.exprs[0], target, decode_registry, path="$")


def parse_many(
    source: str,
    target: Any,
    *,
    registry: DecoderRegistry | None = None,
) -> list[Any]:
    """Parse and decode all top-level Mu expressions as a list."""
    decode_registry = registry or DecoderRegistry()
    doc = parse(source, preserve_spans=True)
    return [
        _decode_value(expr, target, decode_registry, path=f"$[{index}]")
        for index, expr in enumerate(doc.exprs)
    ]


def decode(
    expr: Expr,
    target: Any,
    *,
    registry: DecoderRegistry | None = None,
    path: str = "$",
) -> Any:
    """Decode a pre-parsed Mu `Expr` into a target Python type."""
    decode_registry = registry or DecoderRegistry()
    return _decode_value(expr, target, decode_registry, path)


def _decode_value(
    expr: Expr,
    target: Any,
    registry: DecoderRegistry,
    path: str,
) -> Any:
    target_base, metadata = _unwrap_annotated(target)

    custom = _find_metadata(metadata, DecodeWith)
    if custom is not None:
        return _call_deserializer(custom.fn, expr, target_base, registry, path)

    registry_decoder = registry.get(target_base)
    if registry_decoder is not None:
        return _call_deserializer(registry_decoder, expr, target_base, registry, path)

    return _decode_builtin(expr, target_base, registry, path)


def _decode_builtin(
    expr: Expr,
    target: Any,
    registry: DecoderRegistry,
    path: str,
) -> Any:
    if target is Any:
        return _decode_any(expr, registry, path)

    if isinstance(target, type) and issubclass(target, Expr):
        if isinstance(expr, target):
            return expr
        _raise_decode(path, f"{target.__name__}", expr)

    origin = get_origin(target)

    if origin in {typing.Union, types.UnionType}:
        return _decode_union(expr, list(get_args(target)), registry, path)

    if _is_quoted_type(target):
        args = get_args(target)
        quoted_inner = args[0] if args else Any
        return _decode_quoted(expr, quoted_inner, path)

    if isinstance(target, type) and dataclasses.is_dataclass(target):
        return _decode_dataclass(expr, target, registry, path)

    if target is str:
        if isinstance(expr, StringExpr):
            return expr.value
        if isinstance(expr, AtomExpr):
            return expr.value
        _raise_decode(path, "str", expr)

    if target is bool:
        if not isinstance(expr, AtomExpr):
            _raise_decode(path, "bool (atom true/false)", expr)
        value = expr.value.lower()
        if value == "true":
            return True
        if value == "false":
            return False
        _raise_decode(path, "bool (atom true/false)", expr)

    if target is int:
        if isinstance(expr, SInt):
            return expr.value
        if not isinstance(expr, AtomExpr):
            _raise_decode(path, "int atom", expr)
        try:
            return int(expr.value)
        except ValueError as cause:
            _raise_decode(path, "int atom", expr, cause=cause)

    if target is float:
        if isinstance(expr, SReal):
            return expr.value
        if isinstance(expr, SInt):
            return float(expr.value)
        if isinstance(expr, SRational):
            return expr.value[0] / expr.value[1]
        if not isinstance(expr, AtomExpr):
            _raise_decode(path, "float atom", expr)
        try:
            return float(expr.value)
        except ValueError as cause:
            _raise_decode(path, "float atom", expr, cause=cause)

    if origin is list:
        if not isinstance(expr, SequenceExpr):
            _raise_decode(path, "sequence []", expr)
        args = get_args(target)
        item_type = args[0] if args else Any
        return [
            _decode_value(item, item_type, registry, f"{path}[{index}]")
            for index, item in enumerate(expr.values)
        ]

    if origin is dict:
        if not isinstance(expr, MappingExpr):
            _raise_decode(path, "map {}", expr)
        args = get_args(target)
        key_type = args[0] if len(args) >= 1 else Any
        value_type = args[1] if len(args) >= 2 else Any
        result: dict[Any, Any] = {}
        for index, field in enumerate(expr.values):
            key = _decode_value(field.key, key_type, registry, f"{path}.keys[{index}]")
            value = _decode_value(field.value, value_type, registry, f"{path}.values[{index}]")
            try:
                result[key] = value
            except TypeError as cause:
                _raise_decode(path, "hashable map key", expr, cause=cause)
        return result

    if target is Document:
        _raise_decode(path, "Document is not a valid expression target", expr)

    if isinstance(target, type):
        _raise_decode(path, target.__name__, expr)

    _raise_decode(path, str(target), expr)


def _decode_any(expr: Expr, registry: DecoderRegistry, path: str) -> Any:
    if isinstance(expr, SInt):
        return expr.value
    if isinstance(expr, SReal):
        return expr.value
    if isinstance(expr, SRational):
        return expr.value
    if isinstance(expr, AtomExpr):
        return expr.value
    if isinstance(expr, StringExpr):
        return expr.value
    if isinstance(expr, SequenceExpr):
        return [
            _decode_any(item, registry, f"{path}[{index}]")
            for index, item in enumerate(expr.values)
        ]
    if isinstance(expr, MappingExpr):
        result: dict[Any, Any] = {}
        for index, field in enumerate(expr.values):
            key = _decode_any(field.key, registry, f"{path}.keys[{index}]")
            value = _decode_any(field.value, registry, f"{path}.values[{index}]")
            result[key] = value
        return result
    if isinstance(expr, GroupExpr):
        return [
            _decode_any(item, registry, f"{path}[{index}]")
            for index, item in enumerate(expr.values)
        ]
    return expr


def _decode_union(
    expr: Expr,
    union_members: list[Any],
    registry: DecoderRegistry,
    path: str,
) -> Any:
    members: list[Any] = []
    for member in union_members:
        base, _ = _unwrap_annotated(member)
        members.append(base)

    none_member = type(None)
    has_none = any(member is none_member for member in members)

    dataclass_members = [
        member
        for member in members
        if isinstance(member, type) and dataclasses.is_dataclass(member)
    ]
    if dataclass_members and len(dataclass_members) == len(
        [member for member in members if member is not none_member]
    ):
        return _decode_dataclass_union(expr, dataclass_members, path, registry)

    if has_none and _is_none_atom(expr):
        return None

    errors: list[DecodeError] = []
    for member in members:
        if member is none_member:
            continue
        try:
            return _decode_value(expr, member, registry, path)
        except DecodeError as error:
            errors.append(error)

    if errors:
        _raise_decode(
            path,
            " | ".join(_target_name(member) for member in members),
            expr,
            cause=errors[0],
        )

    _raise_decode(path, "non-empty union", expr)


def _decode_dataclass_union(
    expr: Expr,
    members: list[type[Any]],
    path: str,
    registry: DecoderRegistry,
) -> Any:
    if not isinstance(expr, GroupExpr) or not expr.values:
        _raise_decode(path, "tagged group for dataclass union", expr)

    head = expr.values[0]
    if not isinstance(head, AtomExpr):
        _raise_decode(path, "tag atom", head)

    tag_to_type: dict[str, type[Any]] = {}
    for member in members:
        tag = _dataclass_tag(member)
        if tag in tag_to_type and tag_to_type[tag] is not member:
            raise DecodeError(
                path=path,
                expected="unique dataclass union tags",
                got=f"duplicate tag '{tag}'",
            )
        tag_to_type[tag] = member

    selected = tag_to_type.get(head.value)
    if selected is None:
        expected_tags = ", ".join(sorted(tag_to_type))
        _raise_decode(path, f"one of dataclass tags: {expected_tags}", head)

    return _decode_dataclass(expr, selected, registry, path)


def _decode_dataclass(
    expr: Expr,
    target: type[Any],
    registry: DecoderRegistry,
    path: str,
) -> Any:
    if not isinstance(expr, GroupExpr) or not expr.values:
        _raise_decode(path, f"tagged group for {target.__name__}", expr)

    head = expr.values[0]
    if not isinstance(head, AtomExpr):
        _raise_decode(path, f"tag atom for {target.__name__}", head)

    expected_tag = _dataclass_tag(target)
    if head.value != expected_tag:
        _raise_decode(path, f"tag '{expected_tag}'", head)

    specs = _field_specs(target)
    parameters = [(spec.arity, spec.mu_name) for spec in specs]
    arguments = _group_tail_to_match_args(expr.values[1:])

    try:
        assigned = match_args(parameters, arguments)
    except MatchArgsException as cause:
        _raise_decode(path, f"arguments for {target.__name__}", expr, cause=cause)

    kwargs: dict[str, Any] = {}
    for spec in specs:
        raw_values = assigned.get(spec.mu_name, [])

        if spec.arity == ArgArity.Required:
            if len(raw_values) != 1:
                _raise_decode(path, f"single value for '{spec.mu_name}'", expr)
            kwargs[spec.field_name] = _decode_value(
                raw_values[0],
                spec.type_hint,
                registry,
                f"{path}.{spec.field_name}",
            )
            continue

        if spec.arity == ArgArity.Optional:
            if not raw_values:
                if spec.has_default:
                    continue
                kwargs[spec.field_name] = None
            else:
                kwargs[spec.field_name] = _decode_value(
                    raw_values[0],
                    spec.type_hint,
                    registry,
                    f"{path}.{spec.field_name}",
                )
            continue

        item_type = _list_item_type(spec.base_type)
        kwargs[spec.field_name] = [
            _decode_value(value, item_type, registry, f"{path}.{spec.field_name}[{index}]")
            for index, value in enumerate(raw_values)
        ]

    try:
        return target(**kwargs)
    except Exception as cause:  # pragma: no cover - constructor errors are surfaced as decode errors
        _raise_decode(path, f"construct {target.__name__}", expr, cause=cause)


def _group_tail_to_match_args(
    values: list[Expr],
) -> list[NamedArg[str] | PositionalArg[Expr]]:
    result: list[NamedArg[str] | PositionalArg[Expr]] = []
    for value in values:
        if isinstance(value, AtomExpr) and value.value.startswith(":"):
            result.append(NamedArg(value.value[1:]))
        else:
            result.append(PositionalArg(value))
    return result


def _field_specs(target: type[Any]) -> tuple[_FieldSpec, ...]:
    type_hints = typing.get_type_hints(target, include_extras=True)
    specs: list[_FieldSpec] = []

    for field in dataclasses.fields(target):
        hint = type_hints.get(field.name, Any)
        base_type, metadata = _unwrap_annotated(hint)

        name_meta = _find_metadata(metadata, FieldName)
        mu_name = name_meta.name if name_meta is not None else _snake_to_kebab(field.name)

        marker = _pick_arity_marker(metadata)
        has_default = field.default is not MISSING or field.default_factory is not MISSING

        if marker is OptionalArg:
            arity = ArgArity.Optional
        elif marker is ZeroOrMore:
            arity = ArgArity.ZeroOrMore
        elif marker is OneOrMore:
            arity = ArgArity.OneOrMore
        else:
            arity = ArgArity.Optional if has_default else ArgArity.Required

        if arity in {ArgArity.ZeroOrMore, ArgArity.OneOrMore}:
            origin = get_origin(base_type)
            if origin is not list:
                raise TypeError(
                    f"Field '{field.name}' uses vararg marker but type is not list[T]: {base_type}"
                )

        specs.append(
            _FieldSpec(
                field_name=field.name,
                mu_name=mu_name,
                type_hint=hint,
                base_type=base_type,
                metadata=tuple(metadata),
                arity=arity,
                has_default=has_default,
            )
        )

    return tuple(specs)


def _list_item_type(target: Any) -> Any:
    origin = get_origin(target)
    if origin is not list:
        return Any
    args = get_args(target)
    return args[0] if args else Any


def _unwrap_annotated(target: Any) -> tuple[Any, list[Any]]:
    metadata: list[Any] = []
    current = target
    while get_origin(current) is Annotated:
        args = list(get_args(current))
        if not args:
            break
        current = args[0]
        metadata.extend(args[1:])
    return current, metadata


def _find_metadata(metadata: list[Any], target_type: type[T]) -> T | None:
    found: T | None = None
    for item in metadata:
        if isinstance(item, target_type):
            if found is not None:
                raise TypeError(f"Multiple {target_type.__name__} annotations are not allowed")
            found = item
    return found


def _is_marker(item: Any, marker: type[Any]) -> bool:
    return item is marker or isinstance(item, marker)


def _pick_arity_marker(metadata: list[Any]) -> type[Any] | None:
    markers: list[type[Any]] = []
    if any(_is_marker(item, OptionalArg) for item in metadata):
        markers.append(OptionalArg)
    if any(_is_marker(item, ZeroOrMore) for item in metadata):
        markers.append(ZeroOrMore)
    if any(_is_marker(item, OneOrMore) for item in metadata):
        markers.append(OneOrMore)

    if len(markers) > 1:
        names = ", ".join(marker.__name__ for marker in markers)
        raise TypeError(f"At most one arity marker is allowed, found: {names}")
    return markers[0] if markers else None


def _dataclass_tag(target: type[Any]) -> str:
    override = getattr(target, "__mu_tag__", None)
    if override is not None:
        return str(override)
    return _snake_to_kebab(_camel_to_snake(target.__name__))


def _camel_to_snake(name: str) -> str:
    step1 = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", step1).lower()


def _snake_to_kebab(name: str) -> str:
    return name.replace("_", "-")


def _call_deserializer(
    fn: DecoderFn,
    expr: Expr,
    target: Any,
    registry: DecoderRegistry,
    path: str,
) -> Any:
    context = DecodeContext(path=path, target=target, registry=registry)
    try:
        return fn(expr, context)
    except DecodeError:
        raise
    except Exception as cause:
        _raise_decode(path, f"custom deserializer for {_target_name(target)}", expr, cause=cause)


def _decode_quoted(expr: Expr, inner: Any, path: str) -> Any:
    if inner is str:
        if isinstance(expr, StringExpr):
            return Quoted(expr.value)
        if isinstance(expr, AtomExpr):
            return Quoted(expr.value)
        _raise_decode(path, "Quoted[str]", expr)

    if isinstance(inner, type) and issubclass(inner, Expr):
        if isinstance(expr, inner):
            return Quoted(expr)
        _raise_decode(path, f"Quoted[{inner.__name__}]", expr)

    _raise_decode(path, "Quoted[str] or Quoted[Expr subtype]", expr)


def _is_quoted_type(target: Any) -> bool:
    return get_origin(target) is Quoted


def _target_name(target: Any) -> str:
    if isinstance(target, type):
        return target.__name__
    return str(target)


def _is_none_atom(expr: Expr) -> bool:
    return isinstance(expr, AtomExpr) and expr.value.lower() in {"none", "null"}


def _describe_expr(expr: Any) -> str:
    if isinstance(expr, SInt):
        return f"int({expr.value!r})"
    if isinstance(expr, SReal):
        return f"real({expr.value!r})"
    if isinstance(expr, SRational):
        return f"rational({expr.value[0]!r}/{expr.value[1]!r})"
    if isinstance(expr, AtomExpr):
        return f"atom({expr.value!r})"
    if isinstance(expr, StringExpr):
        return f"string({expr.value!r})"
    if isinstance(expr, GroupExpr):
        return "group(...)"
    if isinstance(expr, SequenceExpr):
        return "seq[...]"
    if isinstance(expr, MappingExpr):
        return "map{...}"
    return type(expr).__name__


def _extract_span(expr: Any) -> Any | None:
    if isinstance(expr, (AtomExpr, StringExpr, SInt, SReal, SRational)):
        span = expr.span
        if span is None:
            return None
        token = getattr(span, "token", None)
        return token if token is not None else span

    for attr in ("open_bracket", "span"):
        value = getattr(expr, attr, None)
        if value is not None:
            token = getattr(value, "token", None)
            return token if token is not None else value

    return None


def _raise_decode(
    path: str,
    expected: str,
    expr: Any,
    cause: Exception | None = None,
) -> typing.NoReturn:
    raise DecodeError(
        path=path,
        expected=expected,
        got=_describe_expr(expr),
        span=_extract_span(expr),
        cause=cause,
    )


__all__ = [
    "DecodeContext",
    "DecodeError",
    "DecodeWith",
    "DecoderFn",
    "DecoderRegistry",
    "FieldName",
    "OneOrMore",
    "OptionalArg",
    "ZeroOrMore",
    "decode",
    "tag",
    "parse_many",
    "parse_one",
]



"""Parser for Mu source text into AST nodes."""

from __future__ import annotations

import re
from math import gcd

from mu.input import Span, _Input, debug
from mu.types import (
    AtomExpr,
    Document,
    Expr,
    GroupExpr,
    MappingExpr,
    MappingField,
    SequenceExpr,
    SInt,
    SRational,
    SReal,
    StringExpr,
    TokenSpans,
)


class ParseError(Exception):
    """Raised when Mu source cannot be parsed into a valid AST."""

    pass


_NAME_EXTRA_CHARS = "_.@/+-$%=!?*#&~^|<>:'"
_INTEGER_RE = re.compile(r"^[+-]?[0-9](?:[0-9_]*[0-9])?$")
_RATIONAL_RE = re.compile(
    r"^(?P<num>[+-]?[0-9](?:[0-9_]*[0-9])?)/(?P<den>[0-9](?:[0-9_]*[0-9])?)$"
)
_REAL_RE = re.compile(
    r"^[+-]?[0-9](?:[0-9_]*[0-9])?\.[0-9](?:[0-9_]*[0-9])?(?:[eE][+-]?[0-9]+)?$"
)
_REAL_DOT_RE = re.compile(r"^[+-]?[0-9](?:[0-9_]*[0-9])?\.(?:[eE][+-]?[0-9]+)?$")
_REAL_EXP_RE = re.compile(
    r"^[+-]?[0-9](?:[0-9_]*[0-9])?(?:\.[0-9](?:[0-9_]*[0-9])?)?[eE][+-]?[0-9]+$"
)
_PERCENT_RE = re.compile(
    r"^(?P<value>[+-]?[0-9](?:[0-9_]*[0-9])?(?:\.[0-9](?:[0-9_]*[0-9])?)?)%$"
)


def _is_name_char(ch: str) -> bool:
    if ch == _Input.EOS:
        return False
    return ch.isalnum() or ch in _NAME_EXTRA_CHARS


def _clean_numeric(text: str) -> str:
    return text.replace("_", "")


@debug
def _parse_one_sexpr(input: _Input) -> Expr:
    _skip_whitespace(input)
    c = input.current
    if c == _Input.EOS:
        raise ParseError("Unexpected end of input")
    if c == "(":
        return _parse_group(input)
    elif c == "[":
        return _parse_list(input)
    elif c in {'"', "'"}:
        return _parse_string(input, quote=c)
    elif c == "#":
        return _parse_raw_string(input)
    elif c == "{":
        return _parse_map(input)
    elif _is_name_char(c):
        return _parse_symbol_or_number(input)
    raise ParseError(f"Expected expression, found {c!r}")


@debug
def _skip_whitespace(input: _Input) -> Span:
    while True:
        if input.current == _Input.EOS:
            break
        if input.current.isspace():
            input.next()
        elif input.current == ";":
            while input.current not in ["\n", _Input.EOS]:
                input.next()
            if input.current != _Input.EOS:
                input.next()  # consume the newline if it actually was a newline
        else:
            break
    return input.capture()


@debug
def _parse_group(input: _Input) -> GroupExpr:
    if input.current != "(":
        raise ParseError(f"Expected '(' but got {input.current!r}")
    input.next()
    open_bracket = TokenSpans(token=input.capture(), space=_skip_whitespace(input))

    values: list[Expr] = []
    separators: list[TokenSpans] = []
    while input.current != ")":
        if input.current == _Input.EOS:
            break

        values.append(_parse_one_sexpr(input))

        if input.current == ",":  # Skip empty values
            input.next()

        separators.append(
            TokenSpans(token=input.capture(), space=_skip_whitespace(input))
        )

    if input.current == _Input.EOS:
        raise ParseError("Unexpected end of input")

    if input.current != ")":
        raise ParseError(f"Expected ')' but got {input.current!r}")
    input.next()
    close_bracket = TokenSpans(token=input.capture(), space=_skip_whitespace(input))

    return GroupExpr(
        values=values,
        open_bracket=open_bracket,
        separators=separators,
        close_bracket=close_bracket,
    )


@debug
def _parse_list(input: _Input) -> SequenceExpr:
    if input.current != "[":
        raise ParseError(f"Expected '[' but got {input.current!r}")
    input.next()
    open_bracket = TokenSpans(token=input.capture(), space=_skip_whitespace(input))

    values: list[Expr] = []
    separators: list[TokenSpans] = []
    while input.current != "]":
        if input.current == _Input.EOS:
            raise ParseError("Unexpected end of input")

        values.append(_parse_one_sexpr(input))

        if input.current == ",":  # Skip empty values
            input.next()

        separators.append(
            TokenSpans(token=input.capture(), space=_skip_whitespace(input))
        )

    if input.current != "]":
        raise ParseError(f"Expected ']' but got {input.current!r}")
    input.next()
    close_bracket = TokenSpans(token=input.capture(), space=_skip_whitespace(input))

    return SequenceExpr(
        values=values,
        open_bracket=open_bracket,
        separators=separators,
        close_bracket=close_bracket,
    )


@debug
def _parse_symbol_or_number(input: _Input) -> Expr:
    if not _is_name_char(input.current):
        raise ParseError(f"Unexpected character: {input.current!r}")

    value = ""
    while _is_name_char(input.current):
        value += input.current
        input.next()

    value_span = TokenSpans(token=input.capture(), space=_skip_whitespace(input))
    numeric_expr = _parse_number_token(value, value_span)
    if numeric_expr is not None:
        return numeric_expr
    return AtomExpr(value=value, span=value_span)


def _parse_number_token(value: str, span: TokenSpans) -> Expr | None:
    rational_match = _RATIONAL_RE.fullmatch(value)
    if rational_match is not None:
        numerator = int(_clean_numeric(rational_match.group("num")))
        denominator = int(_clean_numeric(rational_match.group("den")))
        if denominator == 0:
            raise ParseError("Rational denominator cannot be zero")
        factor = gcd(abs(numerator), denominator)
        return SRational(value=(numerator // factor, denominator // factor), span=span)

    percent_match = _PERCENT_RE.fullmatch(value)
    if percent_match is not None:
        return SReal(value=float(_clean_numeric(percent_match.group("value"))) / 100.0, span=span)

    if _REAL_RE.fullmatch(value) or _REAL_DOT_RE.fullmatch(value) or _REAL_EXP_RE.fullmatch(value):
        return SReal(value=float(_clean_numeric(value)), span=span)

    if _INTEGER_RE.fullmatch(value):
        return SInt(value=int(_clean_numeric(value)), span=span)

    return None


@debug
def _parse_string(input: _Input, quote: str = '"') -> StringExpr:
    if quote not in {'"', "'"}:
        raise ParseError(f"Invalid string quote delimiter {quote!r}")
    if input.current != quote:
        raise ParseError(f"Expected {quote!r} but got {input.current!r}")
    input.next()

    value = ""
    while input.current != quote:
        if input.current == _Input.EOS:
            raise ParseError("Unexpected end of input")
        if input.current == "\\":
            input.next()
            if input.current == _Input.EOS:
                raise ParseError("Unexpected end of input in escape sequence")
            match input.current:
                case "n":
                    value += "\n"
                    input.next()
                case "t":
                    value += "\t"
                    input.next()
                case "r":
                    value += "\r"
                    input.next()
                case "0":
                    value += "\0"
                    input.next()
                case "\\":
                    value += "\\"
                    input.next()
                case "'":
                    value += "'"
                    input.next()
                case '"':
                    value += '"'
                    input.next()
                case "u":
                    input.next()
                    value += _parse_unicode_escape(input)
                case _:
                    # Kotlin parser preserves unknown escapes verbatim.
                    value += "\\"
                    value += input.current
                    input.next()
        else:
            value += input.current
            input.next()

    if input.current != quote:
        raise ParseError(f"Expected closing quote {quote!r}")
    input.next()

    value_span = TokenSpans(token=input.capture(), space=_skip_whitespace(input))

    return StringExpr(value=value, span=value_span)


def _parse_unicode_escape(input: _Input) -> str:
    if input.current != "{":
        raise ParseError("Expected '{' after '\\u' in string escape")
    input.next()

    digits = ""
    while input.current != "}":
        if input.current == _Input.EOS:
            raise ParseError("Unexpected end of input in unicode escape")
        if not input.current.isdigit() and input.current.lower() not in "abcdef":
            raise ParseError("Expected hex digit in unicode escape")
        digits += input.current
        input.next()

    if digits == "":
        raise ParseError("Unicode escape must contain at least one hex digit")

    input.next()

    codepoint = int(digits, 16)
    try:
        return chr(codepoint)
    except ValueError as error:  # pragma: no cover - defensive for invalid code points
        raise ParseError("Invalid unicode code point in escape") from error


@debug
def _parse_raw_string(input: _Input) -> StringExpr:
    if input.current != "#":
        raise ParseError(f"Expected '#' but got {input.current!r}")
    input.next()
    tag = ""
    while input.current.isalnum():
        tag += input.current
        input.next()
    if input.current != '"':
        raise ParseError("Expected '\"' after raw string tag")
    input.next()
    value = ""

    while True:
        if input.current == _Input.EOS:
            raise ParseError("Unexpected end of input")
        if input.current != '"':
            value += input.current
            input.next()
            continue
        else:
            input.next()
            if tag == "":
                break

            count = 0
            while input.current == tag[count]:
                count += 1
                input.next()
                if count == len(tag):
                    break

            if count == len(tag):
                break

            value += '"' + tag[:count]
            continue

    value_span = TokenSpans(token=input.capture(), space=_skip_whitespace(input))

    return StringExpr(value=value, span=value_span)


@debug
def _parse_map(input: _Input) -> MappingExpr:
    if input.current != "{":
        raise ParseError(f"Expected '{{' but got {input.current!r}")
    input.next()
    open_bracket = TokenSpans(token=input.capture(), space=_skip_whitespace(input))

    values: list[MappingField] = []
    separators: list[TokenSpans] = []
    while input.current != "}":
        if input.current == _Input.EOS:
            raise ParseError("Unexpected end of input")

        key = _parse_one_sexpr(input)

        if isinstance(key, AtomExpr) and key.value.endswith(":"):
            # We are gonna have to do some surgery here
            if key.span is None or key.span.token is None:
                raise ParseError("Missing span information for map key")
            key_span = key.span.token
            new_key = AtomExpr(key.value[:-1], TokenSpans(key_span[:-1], key_span[-1:-1]))
            colon_span = key_span[-1:]
            colon = TokenSpans(token=colon_span, space=key.span.space)
            key = new_key
        else:
            if input.current != ":":
                raise ParseError(f"Expected ':' but got '{input.current}'")
            input.next()
            colon = TokenSpans(token=input.capture(), space=_skip_whitespace(input))

        value = _parse_one_sexpr(input)
        values.append(MappingField(key, value, colon))

        if input.current == ",":
            input.next()
        separators.append(
            TokenSpans(token=input.capture(), space=_skip_whitespace(input))
        )

    if input.current != "}":
        raise ParseError(f"Expected '}}' but got {input.current!r}")
    input.next()

    close_bracket = TokenSpans(token=input.capture(), space=_skip_whitespace(input))

    return MappingExpr(
        values=values,
        open_bracket=open_bracket,
        separators=separators,
        close_bracket=close_bracket,
    )


@debug
def parse(
    input: str,
    preserve_spans: bool = False,
) -> Document:
    """Parse Mu source text into a `Document`.

    Args:
        input: Mu source text.
        preserve_spans: When `True`, keep token/space span metadata on nodes.
            Default `False` returns span-free nodes.

    Returns:
        Parsed `Document`.
    """
    top_level: list[Expr] = []
    input_r = _Input(input)
    leading_space = _skip_whitespace(input_r)
    while input_r.current != _Input.EOS:
        top_level.append(_parse_one_sexpr(input_r))

    result = Document(top_level, leading_space=leading_space)
    if not preserve_spans:
        result = result.drop_spans()
    return result



"""Experimental Mu expression evaluation runtime."""

import inspect
import re
import typing
from abc import ABC, abstractmethod
from collections import OrderedDict
from collections.abc import Callable
from dataclasses import dataclass, field
from fractions import Fraction
from typing import Any

from mu.parser import parse
from mu.quoted import Quoted
from mu.types import (
    AtomExpr,
    Document,
    Expr,
    GroupExpr,
    MappingExpr,
    SequenceExpr,
    SInt,
    SRational,
    SReal,
    StringExpr,
)


class EvalNameError(NameError):
    """Raised when evaluating an unbound symbol name."""

    pass


@dataclass
class FunctionSignature:
    arg_names: list[str]
    arg_types: dict[str, Any]
    return_type: Any

    def __str__(self):
        def format_type(t: Any):
            if t == Any:
                return "Any"
            # if t has type arguments
            if hasattr(t, "__args__"):
                return f"{t.__origin__.__name__}[{', '.join([format_type(arg) for arg in t.__args__])}]"
            return t.__name__

        args_str = ", ".join(
            [f"{name}: {format_type(self.arg_types[name])}" for name in self.arg_names]
        )
        return_str = self.return_type.__name__
        return f"({args_str}) -> {return_str}"


class CallableObject(ABC):
    @abstractmethod
    def __call__(self, ctx: "EvalContext", *args: Any, **kwargs: Any) -> Any:
        pass

    @abstractmethod
    def get_signature(self) -> FunctionSignature:
        pass


@dataclass
class NativeFunction(CallableObject):
    fn: Callable
    signature: FunctionSignature

    def __call__(self, ctx: "EvalContext", *args: Any, **kwargs: Any) -> Any:
        bound_args = self.signature.arg_names[: len(args)]

        arg_types = self.signature.arg_types

        def eval_arg_maybe(ctx, arg, type):
            if hasattr(type, "__origin__") and type.__origin__ == Quoted:
                assert isinstance(arg, Expr), f"Expected Expr, got {type(arg)}"
                assert issubclass(
                    type.__args__[0], Expr
                ), f"Expected Expr subtype, got {type.__args__[0]}"
                if not isinstance(arg, type.__args__[0]):
                    raise TypeError(f"Expected {type.__args__[0]}, got {arg}")
                return Quoted(arg)
            return eval_expr(ctx, arg)

        # Evaluate all arguments
        evaluated_args = [
            eval_arg_maybe(ctx, arg, arg_types.get(bound_args[arg_index], Any))
            for arg_index, arg in enumerate(args)
        ]
        evaluated_kwargs = {
            k: eval_arg_maybe(ctx, v, arg_types.get(k, Any)) for k, v in kwargs.items()
        }

        # print(f"Calling {self.fn.__name__} with args: {evaluated_args} and kwargs: {evaluated_kwargs}")

        return self.fn(*evaluated_args, **evaluated_kwargs)

    def get_signature(self) -> FunctionSignature:
        return self.signature


@dataclass
class EvalContext:
    """Evaluation context holding callable/value bindings."""

    env: dict[str, Any] = field(default_factory=dict)

    def register(
        self, func: Callable | None = None, name: str | None = None
    ) -> Callable | NativeFunction:
        name_override = name

        def decorator(f):
            sig = inspect.signature(f)
            arg_names = list(sig.parameters.keys())
            type_hints = typing.get_type_hints(f)

            arg_types = {name: type_hints.get(name, Any) for name in arg_names}
            return_type = type_hints.get("return", Any)

            func_signature = FunctionSignature(arg_names, arg_types, return_type)
            native_func = NativeFunction(f, func_signature)
            name = name_override or f.__name__
            self.env[name] = native_func
            return native_func

        if func is None:
            return decorator
        else:
            return decorator(func)

    def get_function_signature(self, func_name: str) -> str:
        if func_name not in self.env or not isinstance(
            self.env[func_name], CallableObject
        ):
            return f"Function '{func_name}' not found"

        return str(self.env[func_name].get_signature())


def eval_expr(
    ctx: EvalContext,
    e: Document | Expr | list[Expr],
    ignore_toplevel_exceptions: bool = False,
) -> Any:
    """Evaluate Mu expressions using a provided `EvalContext`.

    This runtime is experimental and not covered by stable API guarantees.
    """
    if isinstance(e, Document):
        return [eval_expr(ctx, x) for x in e.exprs]
    if isinstance(e, list):
        return [eval_expr(ctx, x) for x in e]

    assert isinstance(e, Expr), f"Expected Expr, got {type(e)}"

    match e:
        case AtomExpr(a):
            if a.startswith("py."):
                a = a[3:]
                module, attr = a.rsplit("/", 1)
                import importlib

                module_obj = importlib.import_module(module)
                result = getattr(module_obj, attr)
                if callable(result):
                    sig = inspect.signature(result)
                    arg_names = list(sig.parameters.keys())
                    type_hints = typing.get_type_hints(result)
                    arg_types = {name: type_hints.get(name, Any) for name in arg_names}
                    return_type = type_hints.get("return", Any)
                    func_signature = FunctionSignature(
                        arg_names, arg_types, return_type
                    )
                    return NativeFunction(result, func_signature)
                return result

            if a not in ctx.env:
                raise EvalNameError(f"Name '{a}' is not defined")

            return ctx.env[a]

        case SInt(i):
            return i

        case SReal(r):
            return r

        case SRational(v):
            return Fraction(v[0], v[1])

        case StringExpr(s):
            if "$" in s:
                result = ""
                last_end = 0
                for m in re.finditer(r"\$\{([^\}]+)\}|\$\$", s):
                    result += s[last_end : m.start()]
                    if m.group() == "$$":
                        result += "$"
                    else:
                        result += str(eval_expr(ctx, parse(m.group(1)).exprs)[-1])
                    last_end = m.end()

                result += s[last_end:]

                return result

            return s

        case SequenceExpr(s):
            return [eval_expr(ctx, e) for e in s]

        case MappingExpr(s):
            result = OrderedDict()
            for field in s:
                result[eval_expr(ctx, field.key)] = eval_expr(ctx, field.value)
            return result

        case GroupExpr(g):
            assert len(g) > 0, "Empty group"

            try:
                fn = eval_expr(ctx, g[0])
                tail = g[1:]
                args = []
                kwargs = {}
                for index, arg in enumerate(tail):
                    if isinstance(arg, AtomExpr) and arg.value.startswith(":"):
                        key = arg.value[1:]
                        assert index + 1 < len(
                            g
                        ), f"Missing value for keyword argument {key}"
                        assert key not in kwargs, f"Duplicate keyword argument {key}"
                        kwargs[key] = tail[index + 1]
                    elif arg not in kwargs.values():
                        args.append(arg)

                if isinstance(fn, CallableObject):
                    return fn(ctx, *args, **kwargs)
                elif callable(fn):
                    evaluated_args = [eval_expr(ctx, arg) for arg in args]
                    evaluated_kwargs = {
                        k: eval_expr(ctx, v) for k, v in kwargs.items()
                    }
                    return fn(*evaluated_args, **evaluated_kwargs)
                else:
                    raise TypeError(f"Cannot call {fn}")
            except Exception as ex:
                if ignore_toplevel_exceptions:
                    import traceback

                    traceback.print_exc()
                    return ex
                else:
                    raise



"""Load Mu documents from strings, files, and paths."""

from __future__ import annotations

import os
from pathlib import Path
from typing import Any, Protocol, TypeVar, cast, overload

from mu.parser import parse
from mu.typed import DecodeError, DecoderRegistry, decode
from mu.types import Document

T = TypeVar("T")


class _TextReadable(Protocol):
    def read(self, size: int = -1) -> str:
        """Read text data from a file-like object."""


Source = str | os.PathLike[str] | _TextReadable


def _read_source(source: Source, *, encoding: str) -> str:
    if hasattr(source, "read"):
        text = cast(_TextReadable, source).read()
        if not isinstance(text, str):
            raise TypeError("load() source.read() must return str text")
        return text
    return Path(source).read_text(encoding=encoding)


@overload
def loads(
    source: str,
    *,
    type: None = None,
    registry: DecoderRegistry | None = None,
    preserve_spans: bool = False,
) -> Document: ...


@overload
def loads(
    source: str,
    *,
    type: type[T],
    registry: DecoderRegistry | None = None,
    preserve_spans: bool = False,
) -> T: ...


@overload
def loads(
    source: str,
    *,
    type: Any,
    registry: DecoderRegistry | None = None,
    preserve_spans: bool = False,
) -> Any: ...


def loads(
    source: str,
    *,
    type: Any | None = None,
    registry: DecoderRegistry | None = None,
    preserve_spans: bool = False,
) -> Document | Any:
    """Load Mu from a string.

    When `type` is omitted, returns `Document`. When provided, decodes a
    single top-level expression into the target type.
    """
    doc = parse(source, preserve_spans=preserve_spans)
    if type is None:
        return doc

    if len(doc.exprs) != 1:
        raise DecodeError(
            path="$",
            expected="exactly one top-level expression",
            got=f"{len(doc.exprs)} expressions",
        )
    return decode(doc.exprs[0], type, registry=registry, path="$")


@overload
def load(
    source: Source,
    *,
    type: None = None,
    encoding: str = "utf-8",
    registry: DecoderRegistry | None = None,
    preserve_spans: bool = False,
) -> Document: ...


@overload
def load(
    source: Source,
    *,
    type: type[T],
    encoding: str = "utf-8",
    registry: DecoderRegistry | None = None,
    preserve_spans: bool = False,
) -> T: ...


@overload
def load(
    source: Source,
    *,
    type: Any,
    encoding: str = "utf-8",
    registry: DecoderRegistry | None = None,
    preserve_spans: bool = False,
) -> Any: ...


def load(
    source: Source,
    *,
    type: Any | None = None,
    encoding: str = "utf-8",
    registry: DecoderRegistry | None = None,
    preserve_spans: bool = False,
) -> Document | Any:
    """Load Mu from a path or file-like object."""
    text = _read_source(source, encoding=encoding)
    return loads(
        text,
        type=type,
        registry=registry,
        preserve_spans=preserve_spans,
    )


__all__ = ["Source", "load", "loads"]
